(*  Title:      HOL/IOA/IOA2.thy
    Author:     Andrej Bogdanov, Stephen J. Garland, Toh Ne Win

The I/O automata of Lynch and Tuttle, in deterministic post state
format as formulated by Bogdanov.

*)

theory IOA2 = Asig + Hilbert_Choice:

theorem existenceSome [simp]: "(EX a. (f a)) ==> f (SOME a. (f a))"
  apply (auto)
  apply (simp add: someI)
done

(* Remember, executions are in forward order: the head of the list is
the last action *)

types
   ('action) execution = "'action list"
   ('action1, 'actions) actionMap = "'action1 => 'actions"
   ('action,'state) transition =   "('state * 'action * 'state)"
   ('action,'state) enablement =  "'state => 'action => bool"
   ('action,'state) effect = "'state => 'action => 'state"

datatype ('action, 'state) ioa =
  IOA 
     "'action signature"    (* Signature *)
     "'state set"           (* Starts *)
     "('action,'state) enablement"  (* Enablement *)
     "('action,'state) effect"  (* Effects *)

constdefs trans_pre :: "('a, 's) transition => 's"
  "trans_pre t == fst(t)"

constdefs trans_act :: "('a, 's) transition => 'a"
  "trans_act t == fst(snd(t))"

constdefs trans_post :: "('a, 's) transition => 's"
  "trans_post t == snd(snd(t))"

constdefs trans_make :: "'s => 'a => 's => ('a, 's) transition"
  "trans_make == % s a s' . (s, a, s')"

(* IO automata *)

consts
  asig_of    ::"('action,'state)ioa => 'action signature"
primrec
  asig_of_def:
  "asig_of (IOA sig starts enablement effects) = sig"

consts
  starts_of  ::"('action,'state)ioa => 'state set"
primrec
  starts_of_def:
  "starts_of (IOA sig starts enablement effects) = starts"

consts
  enablement_of  ::"('action,'state)ioa => ('action, 'state) enablement"
primrec
  enablement_of_def:
  "enablement_of (IOA sig starts enablement effects) = enablement"

consts
  effects_of  ::"('action,'state)ioa => ('action, 'state) effect"
primrec
  effects_of_def:
  "effects_of (IOA sig starts enablement effects) = effects"

consts
  trans_of   ::"('action,'state)ioa => ('action,'state)transition set"

primrec
  trans_of_def:
  "trans_of (IOA sig starts enablement effects) = 
  {(s, a, s') . enablement s a & s' = effects s a}

  "

consts
  isIOA        ::"('action,'state) ioa => bool"
primrec
  isIOA_def:
  "isIOA (IOA sig starts enablement effects) = (
    is_asig(sig)   &
    ~(starts = {}) &
    (ALL acts::'action. ALL s::'state.
     (acts : internals(sig) --> (enablement s acts)))
  )
"

  (* Executions, schedules, and traces *)

consts
  isExecution :: "
                  ('action, 'state) ioa =>
                  'state =>
                  ('action) execution =>
                  bool"

(* The last result of an execution *)
consts lastOf :: "('action, 'state) ioa =>
                  'state =>
                  ('action) execution =>
                  'state"

primrec 
  lastOf_def:
  "lastOf automaton s [] = s"
  lastOf_def2:
  "lastOf automaton s (act#rest) = 
   effects_of automaton (lastOf automaton s rest) act
  "

primrec
  isExecution_def:
  "isExecution automaton s [] = (
   True

  )"
  isExecution_def2:
  "isExecution automaton s (act#rest) = (
   (enablement_of automaton (lastOf automaton s rest) act) &
   (isExecution automaton s rest)
  )"

(* Reachability - with witness variables *)
constdefs 
  reachableWith :: "('action,'state)ioa => 
                    'state =>
                    'state =>
                    'action execution =>
                    bool"
  "reachableWith aut s s0 alpha == (
   (s0 : starts_of aut) &
   (isExecution aut s0 alpha) &
   (lastOf aut s0 alpha = s)
)"

constdefs
  reachable   :: "('action,'state)ioa => 'state => bool"
  "reachable aut s == 
  (EX s0 . True & (EX alpha . reachableWith aut s s0 alpha))
  "

constdefs
  invariant     :: "[('action,'state)ioa, 'state=>bool] => bool"
 "invariant A P == (!s. reachable A s --> P(s))"

theorem invariantReachable:
 assumes  p0: "invariant aut P"
     and  p1: "reachable aut s"
   shows "P s"
proof -
  have f0: "(!s. reachable aut s --> P(s))"
    apply (insert p0)
    apply (simp add: invariant_def) (* WHY doesn't this work in one step?? *)
  done
  show "?thesis"
    apply (insert f0 p1)
    apply (auto)
  done
qed

(* Get traces of an execution (action seq) *)
constdefs traceOf :: "('action, 'state) ioa => 
                      ('action) execution =>
                      'action list"
  "traceOf aut exec ==
  (let
     pred = % act. (~ (act : internals (asig_of aut)))
   in
     filter pred exec
  )
"

constdefs isExecutionFromStart :: 
  "('action, 'state) ioa =>
   ('action) execution =>
   bool"
  "isExecutionFromStart aut exec ==
   (EX s0 . (s0 : starts_of aut)
          & (isExecution aut s0 exec))

"

constdefs tracesOf ::
  "('action, 'state) ioa => ('action list) set"
  "tracesOf aut ==
   {tr . (EX exec. (isExecutionFromStart aut exec) & (tr = traceOf aut exec))}
  "

constdefs executionsOf ::
  "('action, 'state) ioa => ('action execution) set"
  "executionsOf aut ==
   {exec . isExecutionFromStart aut exec}"


(* Simulation relations *)

constdefs isFwdSim_start ::
  "('actionL, 'stateL) ioa => ('actionU, 'stateU) ioa =>
   ('stateL => 'stateU => bool) =>
    bool"

  "isFwdSim_start autL autU relation == 
   (ALL s0. s0: starts_of autL --> (EX u0 . (u0 : starts_of autU) &
                                            (relation s0 u0)
                                   ))
  "

theorem isFwdSim_startRule1:
  assumes p0: "ALL s0. s0: starts_of autL --> (EX u0 . (u0 : starts_of autU) & (relation s0 u0))"
    shows "isFwdSim_start autL autU relation"
  apply (simp add: isFwdSim_start_def)
  apply (simp add: p0)
done

theorem isFwdSim_startRule:
  assumes p0: "!! s0. (s0: starts_of autL ==> (EX u0 . (u0 : starts_of autU) & (relation s0 u0)))"
    shows "isFwdSim_start autL autU relation"
  apply (rule isFwdSim_startRule1)
  apply (simp add: p0)
done

constdefs correspExec ::
  "('actionL, 'stateL) ioa => ('actionU, 'stateU) ioa =>
   ('stateL => 'stateU => bool) =>
   ('actionL, 'action) actionMap =>
   ('actionU, 'action) actionMap =>
   'stateL => 'stateU => 
   ('actionL execution) => ('actionU execution) =>
    bool"
  "correspExec autL autU relation mapL mapU s u alpha beta == True
     & (map mapU (traceOf autU beta))
        = (map mapL (traceOf autL alpha))
     & (isExecution autU u beta)
     & (relation (lastOf autL s alpha) (lastOf autU u beta))
"

constdefs isFwdSim_trans ::
  "('actionL, 'stateL) ioa => ('actionU, 'stateU) ioa =>
   ('stateL => 'stateU => bool) =>
   ('actionL, 'action) actionMap =>
   ('actionU, 'action) actionMap =>
    bool"
  "isFwdSim_trans autL autU relation mapL mapU ==
   (ALL s act u. (enablement_of autL s act) & (relation s u) & (reachable autL s) -->
   (EX beta . correspExec autL autU relation mapL mapU s u [act] beta
   )
  
   )
  "

theorem isFwdSim_transRule:
  assumes p0: "!! sImpl act sSpec. [| (enablement_of autL sImpl act); (relation sImpl sSpec); (reachable autL sImpl) |]
  ==>    (EX beta . correspExec autL autU relation mapL mapU sImpl sSpec [act] beta)
  "
  shows  "isFwdSim_trans autL autU relation mapL mapU"
  apply (simp add: isFwdSim_trans_def)
  apply (auto)
  apply (simp add: p0)
done


constdefs isFwdSim :: "('actionL, 'stateL) ioa => ('actionU, 'stateU) ioa =>
                       ('stateL => 'stateU => bool) =>
                       ('actionL => 'action) =>
                       ('actionU => 'action) =>
                       bool"
  "isFwdSim autL autU relation mapL mapU == 
      (isFwdSim_start autL autU relation)
    & (isFwdSim_trans autL autU relation mapL mapU)
  "

(* theorem executionStep1: *)
(*  "isExecution automaton s (act#rest) ==> isExecution automaton (effects_of automaton s act) rest" *)
(*   apply (simp add: isExecution_def) *)
(* done *)

theorem executionStep:
 assumes a0: "isExecution automaton s alpha"
     and a1: "   enablement_of automaton (lastOf automaton s alpha) act"
    shows "isExecution automaton s (act#alpha)"
  apply (simp add: isExecution_def2 prems)
done

theorem lastOfStep:
 "lastOf automaton s (act#alpha) =
  effects_of automaton (lastOf automaton s alpha) act"
  apply (simp add: lastOf_def)
done

theorem reachableStep:
  "[|
   reachable automaton s;
   enablement_of automaton s act
   |] ==>
   reachable automaton (effects_of automaton s act)"
  apply (simp add: reachable_def)
  proof -
    assume
      a0: "EX s0. Ex (reachableWith automaton s s0)"
      and a1: "enablement_of automaton s act"
  def s0 == "SOME s0 . Ex (reachableWith automaton s s0)"
  have t0: "Ex (reachableWith automaton s s0)"
    apply (simp add: s0_def a0 existenceSome)
    apply (rule existenceSome)
    apply (simp add: a0)
  done
  have t1: "EX u. reachableWith automaton s s0 u"
    apply (simp add: t0)
  done
  def alpha == "SOME u . reachableWith automaton s s0 u"
  have t2: "reachableWith automaton s s0 alpha"
    apply (simp add: alpha_def t1)
  done
  have a2: "s0 : starts_of automaton"
   and a3: "isExecution automaton s0 alpha"
   and a4: "lastOf automaton s0 alpha = s"
    apply (insert t2)  (* Why couldn't I add both to simp?? *)
    apply (simp_all add: reachableWith_def)
  done
  show "EX s0. Ex (reachableWith automaton (effects_of automaton s act) s0)"
    proof (rule exI)+
  show "reachableWith automaton (effects_of automaton s act) s0 (act#alpha)"
    apply (insert t2 a1)
    apply (simp add: reachableWith_def)
  done
  qed
qed

constdefs
  invariant_start :: "[('action,'state)ioa, 'state=>bool] => bool"
  "invariant_start aut I ==
   ALL state . (state : (starts_of aut) --> I state)"

constdefs
  invariant_trans :: "[('action,'state)ioa, 'state=>bool] => bool"
  "invariant_trans aut I ==
   ALL state act . 
      ((reachable aut state) &
       (I state) &
       (enablement_of aut state act)
      )-->
       I(effects_of aut state act)"

theorem executionStep_back:
  "isExecution aut s0 (act#alpha) --> isExecution aut s0 alpha"
  apply (auto)
done

theorem invariantI2:
  assumes p0: "invariant_start aut I"
      and p1: "invariant_trans aut I"
      and p2: "s0 : (starts_of aut)"
     shows
     "isExecution aut s0 alpha --> I (lastOf aut s0 alpha)"
proof (induct alpha)
have f1: "I s0"
proof -
  have f1_1: "ALL state. (state : starts_of aut --> I state)"
    by (insert p0, simp add: invariant_start_def)
      (* Why does it have to be in this order?? *)
  show "?thesis"
    by (insert f1_1, simp add: p2)
qed
{case Nil
  show "?case"
  by (simp add: lastOf_def f1)
}
{case Cons
  show "?case"
  proof (safe)
  assume a0: "isExecution aut s0 (a_ # list_)"
  show "I (lastOf aut s0 (a_ # list_))"
  proof -
  have
      a1: "enablement_of aut (lastOf aut s0 list_) a_"
  and a2: "isExecution aut s0 list_"
    apply (insert a0)
    apply (auto) (* Why doesn't simp work? *)
  done
  have f2: "I (lastOf aut s0 list_)"
    apply (simp add: Cons a2)
  done
  have f3: "reachable aut (lastOf aut s0 list_)"
    apply (simp add: reachable_def)
    apply (rule exI, rule exI)
    proof -
    show "reachableWith aut (lastOf aut s0 list_) s0 list_"
      apply (simp add: reachableWith_def)
      apply (simp add: p2 a1 a2)
    done
  qed
  have f4: "reachable aut (lastOf aut s0 (a_ # list_))"
    apply (simp add: reachable_def)
    apply (rule exI, rule exI)
    proof -
    show "reachableWith aut (effects_of aut (lastOf aut s0 list_) a_) s0 (a_ # list_)"
      apply (simp add: reachableWith_def)
      apply (simp add: p2 a1 a2)
    done
  qed
  have f5: "!! state act.
            [|reachable aut state & I state & enablement_of aut state act|] ==> I (effects_of aut state act)"
    apply (insert p1)
    apply (simp add: invariant_trans_def)
  done
  show "?thesis"
    apply (auto)
    apply (insert f5 f3 f2 a1)
    apply (auto) (* Booyeah, let's see some unification! *)
  done
  qed
  qed
}
qed

theorem invariantI1:
  assumes p0: "invariant_start aut I"
      and p1: "invariant_trans aut I"
      and p2: "s0 : (starts_of aut)"
      and p3: "isExecution aut s0 alpha"
     shows
     "I (lastOf aut s0 alpha)"
  apply (simp add: invariantI2 p0 p1 p2 p3)
done

theorem invariantI:
  assumes p0: "invariant_start aut I"
      and p1: "invariant_trans aut I"
     shows "invariant aut I"
  apply (simp add: invariant_def)
  apply (auto)
  proof -
  fix s
  show "reachable aut s ==> I s  "
    apply (simp add: reachable_def)
  proof - 
  assume a0: "EX s0. Ex (reachableWith aut s s0)"
  show "?thesis"
  proof -
  def s0 == "SOME s0 . Ex (reachableWith aut s s0)"
  have t0: "Ex (reachableWith aut s s0)"
    apply (simp add: s0_def a0 existenceSome)
    apply (rule existenceSome)
    apply (simp add: a0)
  done
  have t1: "EX u. reachableWith aut s s0 u"
    apply (simp add: t0)
  done
  def alpha == "SOME u . reachableWith aut s s0 u"
  have f1: "reachableWith aut s s0 alpha"
    apply (simp add: alpha_def t1)
  done
  have f2: "s0 : starts_of aut"
   and f3: "isExecution aut s0 alpha"
   and f4: "lastOf aut s0 alpha = s"
    apply (insert f1)  (* Why couldn't I add both to simp?? *)
    apply (simp_all add: reachableWith_def)
  done
  show "I s"
    apply (insert f4)
    apply (auto)
    apply (rule invariantI1)
    apply (simp_all add: f2 f3 f4 p0 p1)
  done
  qed
  qed
qed


end

